// Program RY2 is intended for use with terminal program KiTTY/PuTTY as YMODEM reciever.
// ver. 1.5

#undef DEBUG
using System;
using System.IO;
using System.Text;

// Buffer structure:
//   0   |  1   |  2 ... 129/1025   | 130/1026
// ------+------+-------------------+---------
// blockS|blockN|128/1024 byte data | sum

public class ry2
{
	const char NUL = '\x00';
	const int SOH = 0x01;
	const int STX = 0x02;
//	const char EOT = '\x04';
	const char ACK = '\x06';
	const char NAK = '\x15';
	const int  CAN = 0x18;

	public static	int  hStart;
	public static	int  blockLen;
	public static	char[] inputBuffer = new char[1027];
	public static	byte[] buffer;
	public static	byte counter;
	public static	FileStream  workFile;
	public static string fileName;


	#if DEBUG
	//	public static StreamWriter log = new StreamWriter("c:\\pp\\ry2.log", false, Encoding.Default);
	public static StreamWriter log = new StreamWriter("c:\\Users\\oem\\Documents\\kitty\\ry2.log", false, Encoding.Default);
	#endif

	public static void Main()
	{
		// init communication
		Console.Write(NAK);

		while (RecieveZeroBlock()) {
			StringBuilder fn = new StringBuilder(12);
			for (int i = 2; inputBuffer[i] != NUL; i++) {
				fn.Append(inputBuffer[i]);
			}
			fileName = fn.ToString();
			workFile = new FileStream(fileName, FileMode.Create);
			while (ReadHeader()) {
				RecieveBlock(true);
			}
			// EOT recieved
			workFile.Close();
			Console.Write(NAK);
			Console.Read(); // EOT
			Console.Write(ACK);
		}
		#if DEBUG
		log.Close();
		#endif
	}


	private static bool RecieveZeroBlock()
	{
		counter = 0;
		ReadHeader(); // SOH or STX
		RecieveBlock(false);
		return buffer[2] != 0;
	}


	private static bool ReadHeader()
	{
	  hStart = Console.Read();
	  if (hStart == CAN) {
			workFile.Close();
			#if DEBUG
			log.Close();
			#endif
			Environment.Exit(1);
		}
	  blockLen = hStart == SOH ? 128 : 1024;
	  return hStart == SOH || hStart == STX;
	}


	private static void RecieveBlock(bool toFile)
	{
		// read block into buffer, data are encoded by Console class
		for (int i = 0; i < blockLen + 3; i++) {
			inputBuffer[i] = (char)Console.Read();
		}
		// convert back to original
		buffer = Console.InputEncoding.GetBytes(inputBuffer);
		byte blockS = buffer[0]; // block No
		byte blockN = buffer[1]; // inverted block No
		byte sum = buffer[blockLen + 2];  // low byte of checksum
		byte sumI = 0;           // internal checksum
		for (int i = 2; i < blockLen + 2; i++) {
			sumI += buffer[i];
		}
		#if DEBUG
		log.WriteLine("block: {0}, Nblock: {1}, hStart: {2}",
			blockS,
			blockN,
			hStart
		);
		#endif
		if (blockS == counter && (blockS + blockN) == 255 && sum == sumI) {
			// block is OK
			// write data into file
			if (toFile) {
				workFile.Write(buffer, 2, blockLen);
			}
			counter++;
			Console.Write(ACK);
			#if DEBUG
			log.WriteLine("block OK\n");
			#endif
		} else {
			// wrong block
			Console.Write(NAK);
			#if DEBUG
			log.WriteLine("block wrong\n");
			#endif
		}
	}

}
