// Program SY2 is intended for use with terminal program KiTTY/PuTTY as YMODEM sender.
// ver. 1.5

#undef DEBUG
using System;
using System.IO;
using System.Text;
using System.Runtime.InteropServices;

// Buffer structure:
//   0   |  1   |  2 ... 129/1025   | 130/1026
// ------+------+-------------------+---------
// blockS|blockN|128/1024 byte data | sum

public class sy2
{
	const char SOH = '\x01';
	const char STX = '\x02';
	const char EOT = '\x04';
	const int  ACK = 0x06;
	const int  NAK = 0x15;
	const int  CAN = 0x18;
	const byte EOF = 0x1a;
	const int  RecLength = 128;

	static int DataLength;
	static byte[] Buffer = new byte[1027];
	static int Answer;
	static FileStream WorkFile;
	static int FileRec;
	static char HByte;
	static bool Disable1K = false;

	#if DEBUG
	//	public static StreamWriter log = new StreamWriter("c:\\pp\\sy2.log", false, Encoding.Default);
	public static StreamWriter log = new StreamWriter("c:\\Users\\oem\\Documents\\kitty\\sy2.log", false, Encoding.Default);
	#endif

	public static void Main(string[] args)
	{
		int firstArg = 0;
		if (args[0].ToUpper() == "-1K") {
			Disable1K = true;
			firstArg = 1;
		}
		for (int i = firstArg; i < args.Length; i++) {
			string fileName = args[i];
			if (Path.GetExtension(fileName).ToUpper() == ".YFL") {
				foreach (string fn in File.ReadAllLines(fileName, Encoding.Default)) {
					if (fn.IndexOf('\\') == -1) {
						SendFile(Path.Combine(Path.GetDirectoryName(fileName), fn));
					} else {
						SendFile(fn);
					}
				}
			} else {
				SendFile(fileName);
			}
		}
		// end of batch
		ZeroBlock();
		SendBlock();

		#if DEBUG
		log.Close();
		#endif
	} // Main


	private static void Abort()
	{
		WorkFile.Close();
		#if DEBUG
		log.Close();
		#endif
		Environment.Exit(1);
	}

	private static void SendFile(string fileName)
	{
		// Zero block with filename
		ZeroBlock();
		WriteFileName(GetCpmName(fileName));
		SendBlock();
		WorkFile = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read);
		long fileLength = new FileInfo(fileName).Length;
		FileRec = (int)(fileLength / RecLength);
		if (fileLength % RecLength > 0) {
			FileRec++;
		}

		while (FileRec > 0) {
			switch (Answer) {
				case ACK:
					// send new block
					NextBlock();
					break;
				case NAK:
					// send again
					SendBlock();
					break;
				case CAN:
					Abort();
					break;
			}
		}

		WorkFile.Close();
		// end of file
		Console.Write(EOT);
		Answer = Console.Read(); // read NAK
		#if DEBUG
		log.WriteLine("Read answer1 {0}\n", Answer);
		#endif
		Console.Write(EOT);
		Answer = Console.Read(); // read ACK
		#if DEBUG
		log.WriteLine("Read answer2 {0}\n", Answer);
		#endif
	}


	private static void Blk1Rec()
	{
		DataLength = RecLength;
		HByte = SOH;
	}


	private static void Blk8Rec()
	{
		DataLength = RecLength * 8;
		HByte = STX;
	}


	private static void NextBlock()
	{
		Buffer[0]++;
		Buffer[1]--;
		if (Disable1K || FileRec < 8) {
			Blk1Rec();
			FileRec -= 1;
		} else {
			Blk8Rec();
			FileRec -= 8;
		}
		// fill data part of Buffer by EOF according CP/M convention
		FillBlock(EOF);
		WorkFile.Read(Buffer, 2, DataLength);
		SendBlock();
	}


	private static void SendBlock()
	{
		byte sum = 0;
		for (int i = 2; i < DataLength + 2; i++) {
			sum += Buffer[i];
		}
		Buffer[DataLength + 2] = sum;

		// pre-converting to eliminate default output convertion of Console class
		// long way to transfer chr(254) as chr(254) :-)
		byte[] unicodeBytes = Encoding.Convert(Console.OutputEncoding, Encoding.Unicode, Buffer, 0, DataLength + 3);
		char[] unicodeChars = Encoding.Unicode.GetChars(unicodeBytes);
		Console.Write(HByte);
		for (int i = 0; i <  DataLength + 3; i++) {
			Console.Write(unicodeChars[i]);
		}

		//read ACK/NAK
		Answer = Console.Read();

		#if DEBUG
		log.WriteLine("Sent block# {0}, suma: {1}", Buffer[0], Buffer[DataLength + 2]);
		log.WriteLine("Read answer block {0}\n", Answer);
		#endif
	}


	private static void ZeroBlock()
	{
		Blk1Rec();
		FillBlock(0);
		Buffer[0] = 0;
		Buffer[1] = 255;
	}


	private static void FillBlock(byte b)
	{
		for (int i = 2; i < DataLength + 2; i++) {
			Buffer[i] = b;
		}
	}


	private static void WriteFileName (string fn)
	{
		for (int j = 0; j < fn.Length; j++) {
			Buffer[j + 2] = Convert.ToByte(fn[j]);
		}
	}


	private static string GetCpmName(string longPath)
	{
		StringBuilder shortName = new StringBuilder(255);
		GetShortPathName(@longPath, shortName, shortName.Capacity);
		// underline ('_') causes problems in CP/M filename
		return Path.GetFileName(shortName.ToString().ToUpperInvariant()).Replace('_', '-');
	}


	[DllImport("kernel32.dll", CharSet = CharSet.Auto)]
	private static extern int GetShortPathName(
		[MarshalAs(UnmanagedType.LPTStr)] string path,
		[MarshalAs(UnmanagedType.LPTStr)]	StringBuilder shortPath,
		int shortPathLength
	);

}
